-- Purpose: Construct a simple interpreter for the Mini-Haskell language
-- Author:  Ben Bitwissle e.a.
-- Date:    2008 03 26



------------------------------------
-- Imports: Chars and Finite Sets --
------------------------------------


-- import the standard finite set library of Haskell and make its
-- functions also available under the qualifier "Set". Furthermore,
-- hide the two functions null and map in order to avoid ambiguities
-- with respect to the null and map function from the Prelude.
import qualified Data.Set as S
import Data.Char 


-- ***************************************************************************
-- ** Notes on the usage of this file ****************************************
-- ***************************************************************************

{-

   Terms can now be evaluated stepwise (lazy or eager).
   The functions for a stepwise evaluation are 
      evalStepwiseE 
      evalStepwiseL
  For instance, if you type in 
     evalStepwiseE (termF++termG)
  Hugs outputs
    (%x. x (%y. x y)) (%x. y x) =
    (%x. y x) (%y'. (%x. y x) y') =
    (%x. y x) (%y'. y y') =
    y (%y'. y y')
  And if you type in
    evalStepwiseL (termF++termG)
  Hugs outputs
    (%x. x (%y. x y)) (%x. y x) =
    (%x. y x) (%y'. (%x. y x) y') =
    y (%y'. (%x. y x) y')

  This gives you the possibility to check each reduction step
  step in the eager and lazy evaluation strategy.
--}

-- some lambda terms to play with:
-- (cf. http://en.wikipedia.org/wiki/Lambda_calculus#Arithmetic_in_lambda_calculus)

s = "(%x y z. x z (y z))"
k = "(%x y. x)"
i = "(%x. x)"

add = "(%x y s z. x s (y s z))"
mul = "(%x y s z. x (y s) z)"
mul' = "(%x y s. x (y s))"   -- eta-conversion; i.e. partial application
exp' = "(%x y. y x)"

zero  = "(%s z. z)"
one   = "(%s z. s (z))"
two   = "(%s z. s (s z))"
three = "(%s z. s (s (s z)))"

-- example: calculating 2 + 3 in the lambda calculus:
--
--   evalStepwiseE (add++two++three) ~~>
--     (%x y s z. x s (y s z)) (%s z. s (s z)) (%s z. s (s (s z))) =
--     (%y s z. (%s z. s (s z)) s (y s z)) (%s z. s (s (s z))) =
--     (%y s z. (%z. s (s z)) (y s z)) (%s z. s (s (s z))) =
--     (%y s z. s (s (y s z))) (%s z. s (s (s z))) =
--     %s z. s (s ((%s z. s (s (s z))) s z)) =
--     %s z. s (s ((%z. s (s (s z))) z)) =
--     %s z. s (s (s (s (s z))))
--
--   couting the number of 's' shows us that 2+3 is indeed 5 ;-)

mini_true = "(%t f. t)"
mini_false = "(%t f. f)"

mini_and = "(%p q. p q p)"
mini_or  = "(%p q. p p q)"
mini_not = "(%p t f. p f t)"
mini_if  = "(%p.p)"

iszero = "(% n. n (%x. (%t f. f)) (%t f. t))"

--------------------------------------------------
-- the terms needed for the last exercise sheet --
--------------------------------------------------
  
termF = "(%x. x (%y. x y))"
termG = "(%x. y x)"
termS = "(%x y z. x z (y z))"
termK = "(%x y. x)"
termC = "(%x f g. f (g x))"

-- Some more exercises:
--
-- try to do this first by hand and check then using:
--    for lazy:   evalL ex1
--                evalStepwiseL ex1
--    for eager:  evalE ex1
--                evalStepwiseE ex1

ex1  = "((%x y. y x z) z) (%y z. (%x. y y) z)"
ex15 =  "(%x y z. (%a b. b a c) b (%x. y z))"
ex2  = "(%x y z. x ((%a x. y x) b x)) (%x. x) (%b. (%z. b z) a)"




-- ***************************************************************************
-- ** PARSER COMBINATORS *****************************************************
-- ***************************************************************************


-- NOTE: The development of some combinator library for parsers is not
-- difficult per se. However, making a production quality combinator
-- library involves a lot of thinking and optimization. This has already
-- been done as part of the "Parsec" library. In a recent GHC distribution
-- this library can be used by
--
--   import Text.ParserCombinators.Parsec
--
-- for its documentation ask Hoogle with "Parsec"



-- parsers are encapsulated in a new datatype in order to 
-- guarantee type-safety (and to enable making it a Monad instance)
data Parser a = Prs (String -> [(a,String)])

-- apply a parser to a string
parse :: Parser a -> String -> [(a,String)]
parse (Prs p) inp = p inp

-- apply a parser and return the value of the first complete parse
-- NOTE: We could use Maybe to defer error handling
completeParse :: Parser a -> String -> a
completeParse p inp 
  | null results = error "Parse unsuccessful"
  | otherwise    = head results
  where results = [val | (val,[]) <- parse p inp]

-- In order to support the "do"-notation, we have to make
-- our Parser type an instance of the Monad typeclass
instance Monad Parser where
    -- return :: a -> Parser a 
    --
    -- Do not consume any input and return a fixed result.
    return x = Prs (\inp -> [(x,inp)])

    -- (>>=) :: Parser a -> (a -> Parser b) -> Parser b
    --
    -- Sequence a parser with a parser generator. The parser
    -- generator generates a parser depending on the outcome
    -- of the first parser. The generated parser is then applied
    -- to the input.
    p >>= gen = Prs (\inp -> [ (val',out') | (val,out) <- parse p inp, 
                                             (val',out') <- parse (gen val) out ])
    
    -- fail :: String -> Parser a
    --
    -- every monad supports the method fail, which signals an error
    -- described by a string. As we cannot store an error value in
    -- our parser type, we just ignore it and signal a simple failure.
    fail = const failure


--------------------------------------------------------------------------

-- choice combinator
(|||) :: Parser a -> Parser a -> Parser a
p ||| q = Prs (\inp -> (parse p inp) ++ (parse q inp))

-- repetition
many :: Parser a -> Parser [a]    -- 0 or more repetitions of p
many p = many1 p ||| return []    

many1 :: Parser a -> Parser [a]
many1 p = do val <- p               -- 1 or more repetitions of p
             vals <- many p
             return (val:vals)

--------------------------------------------------------------------------

-- some simple parsers
failure = Prs (\inp -> [])          -- the parser that always fails

item :: Parser Char
item = Prs (\inp -> case inp of
                    "" -> []
                    (x:xs) -> [(x,xs)])

sat  :: (Char -> Bool) -> Parser Char
sat p = do x <- item 
           if p x then return x else failure

char  :: Char -> Parser Char
char x = sat (==x)

string       :: String -> Parser String
string ""     = return ""
string (x:xs) = do char x 
                   string xs 
                   return (x:xs)


-- some auxiliary parsers
spaces = many (sat (==' '))

identifier = do spaces
                id <- many1 (sat isIdentLetter)
                spaces
                return id
             where isIdentLetter c = isLetter c || (c == '\'')

token s = do spaces
             t <- string s
             spaces
             return t


-- **************************************************************************
-- ** MINI-HASKELL **********************************************************
-- **************************************************************************

-----------------------------------------------------------------------------
-- The term structure -------------------------------------------------------
-----------------------------------------------------------------------------

data Term = Id String | Ap Term Term | Lam String Term  
            deriving (Show, Eq)

-- testing the structure of our terms
isId (Id _) = True
isId _      = False

isLam (Lam _ _) = True
isLam _         = False

-----------------------------------------------------------------------------
-- Pretty Printing our lambda expressions -----------------------------------
-----------------------------------------------------------------------------

-- for comparison: the old simple pretty printer
prettyPrint' (Id s)    = s
prettyPrint' (Lam s t) = "(%" ++ s ++ ". " ++ (prettyPrint' t) ++ ")"
prettyPrint' (Ap s t)  = "(" ++ (prettyPrint' s) ++ " " ++ (prettyPrint' t) ++ ")"

-- We adapt this pretty printer now such that as few as possible '%' and
-- parentheses are printed.
    
-- helper function: prints parentheses around the term t if cond is true
parens :: Bool -> String -> String
parens cond t = if cond then "("++t++")" else t

-- a nice pretty print function for our lambda expressions
prettyPrint (Id x)   = x
prettyPrint (Ap x y) = left++" "++right
  where left  = parens (isLam x)      (prettyPrint x)
        right = parens (not (isId y)) (prettyPrint y)

prettyPrint (Lam x y) = left++right y
  where left = "%" ++ x
        right (Lam _ _) = " " ++ tail (prettyPrint y) -- remove "%" and append
        right _         = ". " ++ prettyPrint y

-- pretty prints a sequence of expressions as a series of equivalences
prettyPrintSequence xs = foldl1 (\x y -> x ++ " = \n" ++ y) (map prettyPrint xs)


-----------------------------------------------------------------------------
-- Parsing Lambda Terms -----------------------------------------------------
-----------------------------------------------------------------------------
--
-- grammar:
--   Term  ::= Ident Term* | Lamb Term* | Paren Term*      
--   Paren ::= ( Term )
--   Lamb  ::= % Ident+ . Term  


atom  = ident ||| lamb ||| paren

ident = do id <- identifier 
           return (Id id)

term  = do t <- atom                 -- left associative apply
           ts <- many atom
           return (foldl Ap t ts)

lamb  = do token "%"
           ids <- many1 identifier
           token "."
           t <- term
           return (foldr Lam t ids)

paren = do token "(" 
           t <- term 
           token ")"
           return t

str2term = completeParse term


-----------------------------------------------------------------------------
-- Evaluating Lambda Terms --------------------------------------------------
-----------------------------------------------------------------------------

-- free variables
free :: Term -> S.Set String
free (Id v)    = S.singleton v
free (Ap s t)  = S.union (free s) (free t)
free (Lam v t) = S.delete v (free t)

-- substitution: subst t v s = t[s/v]
subst :: Term -> String -> Term -> Term 
subst (Id x) v s     = if x == v then s else Id x
subst (Ap t1 t2) v s = Ap (subst t1 v s) (subst t2 v s)
subst (Lam x t) v s 
  | x == v                   = Lam x t
  | x `S.notMember` (free s) = Lam x (subst t v s)
  | otherwise                = Lam z (subst (subst t x (Id z)) v s)
  where z       = fresh (S.union (free t) (free s))
        fresh s = (S.fold max "" s)  ++ "'" -- fresh returns an id not in s


-- beta-reduction
beta (Lam x t) t' = subst t x t'


-- lazy evaluation
lazy :: Term -> Term
lazy (Id x)    = (Id x)
lazy (Ap t t') = case r of
  (Lam _ _) -> lazy (beta r t')
  (_)       -> Ap r t' 
  where r = lazy t
lazy t         = t      -- no evaluation under a lambda abstraction


-- eager evaluation
-- Notice that we DO evaluate inside of lambda-abstractions.   
-- That means, all redicies are reduced.
eager :: Term -> Term
eager (Id x)    = (Id x)
eager (Ap t t') = case r of
  (Lam _ _) -> eager (beta r r')
  (_)       -> Ap r r'
  where r  = eager t
        r' = eager t'
eager (Lam x t) = Lam x (eager t) 


evalE = prettyPrint . eager . str2term
evalL = prettyPrint . lazy  . str2term

-------------------------
-- STEPWISE EVALUATION --
-------------------------

-- a stepwise evaluation of terms with the lazy evaluation strategy
lazyStepwise :: Term -> [Term]
lazyStepwise (Id x) = [Id x]
lazyStepwise (Ap t1 t2)  =  
      [Ap x t2 | x <- r1] ++
      (if isLam r1res then lazyStepwise (beta r1res t2) else [])
      where r1 = lazyStepwise t1
            r1res = last r1
lazyStepwise t = [t]      -- lambda abstraction case


-- evaluates a given Term to a list of Terms that represent the stepwise 
-- beta-reductions carried out by the eager evaluation strategy

eagerStepwise :: Term -> [Term]
eagerStepwise (Id x) = [Id x]
eagerStepwise (Ap t1 t2)  = 
       [Ap x t2 | x <- r1] ++ [Ap r1res x | x <- tail r2] ++
       (if isLam r1res then eagerStepwise (beta r1res r2res) else [])
      where r1 = eagerStepwise t1
            r2 = eagerStepwise t2
            r1res = last r1
            r2res = last r2
eagerStepwise (Lam x t) = map (Lam x) (eagerStepwise t)

-- clean stepwise output
evalStepwiseE = putStrLn . prettyPrintSequence . eagerStepwise . str2term
evalStepwiseL = putStrLn . prettyPrintSequence . lazyStepwise  . str2term


--- for tests
hadd = \x y s z -> x s (y s z) 
data Nat = Zero | Succ Nat deriving (Show,Ord,Eq) 



------------------------------------------- =ox ------------- :-P --------------
-- End of File ----------------- :-) ------------------------------- =o) -------
-------------------- :-() ---- 8-) ----------------- ;-)  (-8 ------------------




