---
--- monadic interpreters a la Wadler
--- see Phil Wadler's paper "The essence of functional programming" POPL 92
--- using a state monad to count the evaluation steps
---

{-# LANGUAGE FlexibleInstances #-}

---
--- State monad
---
data StM s a = St (s -> (a, s))

runStM :: StM s a -> s -> (a, s)
runStM (St m) = m

instance Monad (StM s) where
  return x = St (\s -> (x, s))
  m >>= k  = St (\s -> let (a, t) = runStM m s 
                       in runStM (k a) t)
            
-- monad-specific operations                  
get :: StM s s
get = St (\s -> (s, s))

put :: s -> StM s () 
put t = St (\s -> ((), t))


--- 
--- data types for interpreter
---
type Name = String

data Term = Var Name
          | Lit Int
          | Lam Name Term 
          | App Term Term 
          | Add Term Term 
            
data Value = Wrong 
           | Num Int
           | Fun (Value -> StM Int Value)
             
type Environment = [(Name, Value)]


-- show function for values and monad

instance Show Value where
  show Wrong = "<wrong>"
  show (Num i) = show i
  show (Fun _) = "<function>"

instance Show a => Show (StM Int a) where  -- requires -XFlexibleInstances
  show m = let (a, c) = runStM m 0 
           in "Value: " ++ show a ++ "; " ++
              "Count: " ++ show c ++ "."

---
--- interpreter 
---
eval  :: Term -> Environment -> StM Int Value
eval (Var v)   e = lookup_ v e
eval (Lit x)   e = return (Num x)
eval (Lam x t) e = return (Fun (\a -> eval t ((x, a):e)))
eval (App t u) e = do tick               -- added: tick
                      f <- eval t e
                      a <- eval u e
                      apply f a
eval (Add t u) e = do tick               -- added: tick
                      a <- eval t e
                      b <- eval u e
                      add a b

run :: Term -> StM Int Value
run t = eval t []


--- auxiliary function 

tick :: StM Int ()
tick = do c <- get; put (c + 1)

lookup_  :: Name -> Environment -> StM Int Value
lookup_ x []          = return Wrong 
lookup_ x ((y, v):es) = if x == y then return v else lookup_ x es

add :: Value -> Value -> StM Int Value
add (Num x) (Num y) = return (Num (x + y))
add _       _       = return Wrong

apply :: Value -> Value -> StM Int Value
apply (Fun k) a = k a
apply _       _ = return Wrong


-- terms for testing

-- (\x. x + x) (10 + 11)
t0 = (App (Lam "x" (Add (Var "x") (Var "x"))) (Add (Lit 10) (Lit 11)))

-- (\x. x + 1) ((\x. x + x) (10 + 11))
t1 = (App (Lam "x" (Add (Var "x") (Lit 1))) t0)



