---
--- monadic interpreters a la Wadler
--- see Phil Wadler's paper "The essence of functional programming" POPL 92
---

--
-- Exception Monad
--
data ExcM e a = Success a | Exception e

instance Monad (ExcM e) where
  return x = Success x
  (Success a) >>= k = k a
  (Exception e) >>= k = Exception e
  
-- show as instance of Show
instance (Show a, Show e) => Show (ExcM e a) where
  show (Success a) = show a
  show (Exception e) = show e

-- monad-specific operations
throw :: e -> ExcM e a
throw e = Exception e

catch :: ExcM e a -> (e -> ExcM e a) -> ExcM e a
catch (Success a) h   = (Success a) 
catch (Exception e) h = h e

---
--- data types for interpreter
---
type Name = String

data Term = Var Name
          | Lit Int
          | Lam Name Term 
          | App Term Term 
          | Add Term Term             

data Value = Num Int                           -- removed: Wrong 
           | Fun (Value -> ExcM String Value)
             
type Environment = [(Name, Value)]


instance Show Value where
  show (Num i) = show i
  show (Fun _) = "<function>"


--- interpreter
--- no change

eval  :: Term -> Environment -> ExcM String Value
eval (Var v)   e = lookup_ v e
eval (Lit x)   e = return (Num x)
eval (Lam x t) e = return (Fun (\a -> eval t ((x, a):e)))
eval (App t u) e = do f <- eval t e
                      a <- eval u e
                      apply f a
eval (Add t u) e = do a <- eval t e
                      b <- eval u e
                      add a b

run :: Term -> ExcM String Value
run t = eval t []

-- auxiliary functions
-- replaced all occurrences of "return Wrong" by "throw (<error message>)"

lookup_  :: Name -> Environment -> ExcM String Value
lookup_ x []          = throw ("Unbound variable: "++x) 
lookup_ x ((y, v):es) = if x == y then return v else lookup_ x es

apply :: Value -> Value -> ExcM String Value
apply (Fun k) a = k a
apply v       _ = throw ("Not a function: "++show v)

add :: Value -> Value -> ExcM String Value
add (Num x) (Num y) = return (Num (x + y))
add v       w       = throw ("Not a number: "++show v++" or "++show w)


-- terms for testing

test = [t0, t1, t2, t3]

-- (\x. x + x) (10 + 11)
t0 = (App (Lam "x" (Add (Var "x") (Var "x"))) (Add (Lit 10) (Lit 11)))

-- unbound variable
t1 = (Add (Var "x") (Lit 11))

-- not a number
t2 = (Add (Lam "x" (Var "x")) (Lit 11))

-- not a function
t3 = (App (Lit 10) (Lit 11))
