---
--- monadic interpreters a la Wadler
--- see Phil Wadler's paper "The essence of functional programming" POPL 92
---

---
--- Output monad
---
data OutM a = Out (a, String)

runOutM :: OutM a -> (a, String)
runOutM (Out z) = z

instance Monad OutM where
  return x = Out (x, "")
  m >>= k = let (a, r) = runOutM m  
                (b, s) = runOutM (k a)
            in Out (b, r ++ s)

out :: Show a => a -> OutM a
out a = Out (a, show a ++ "; ")

instance Show a => Show (OutM a) where
  show (Out (a, s)) = "Output: " ++ s ++ "\nValue: " ++ show a

---
--- data types for interpreter
---
type Name = String

data Term = Var Name
          | Lit Int
          | Lam Name Term 
          | App Term Term 
          | Add Term Term 
          | Show Term         -- added 
            
data Value = Wrong 
           | Num Int
           | Fun (Value -> OutM Value)
             
type Environment = [(Name, Value)]

instance Show Value where
  show Wrong = "<wrong>"
  show (Num i) = show i
  show (Fun _) = "<function>"


--- interpreter

eval  :: Term -> Environment -> OutM Value
eval (Var v)   e = lookup_ v e
eval (Lit x)   e = return (Num x)
eval (Lam x t) e = return (Fun (\a -> eval t ((x, a):e)))
eval (App t u) e = do f <- eval t e
                      a <- eval u e
                      apply f a
eval (Add t u) e = do a <- eval t e
                      b <- eval u e
                      add a b
eval (Show t)  e = do r <- eval t e
                      out r

run :: Term -> OutM Value
run t = eval t []


--- auxiliary functions

lookup_  :: Name -> Environment -> OutM Value
lookup_ x []          = return Wrong 
lookup_ x ((y, v):es) = if x == y then return v else lookup_ x es

apply :: Value -> Value -> OutM Value
apply (Fun k) a = k a
apply _       _ = return Wrong

add :: Value -> Value -> OutM Value
add (Num x) (Num y) = return (Num (x + y))
add _       _       = return Wrong


-- terms for testing

sAdd t u = Show (Add t u)

-- (\x. x + x) (10 + 11)
t0 = (App (Lam "x" (sAdd (Var "x") (Var "x"))) (sAdd (Lit 10) (Lit 11)))

-- (((1 + 2) + 3) + 4)
t1 = Add (sAdd (sAdd (Lit 1) (Lit 2)) (Lit 3)) (Lit 4)

-- (1 + (2 + (3 + 4)))
t2 = Add (Lit 1) (sAdd (Lit 2) (sAdd (Lit 3) (Lit 4)))

-- 1 + ((\x. x + x) (10 + 11))
t3 = (Add (Lit 1) (Show t0))

